//  Brian Marshall      bmarshal@agt.net
//......................................

//----------------------------------------------------------
//  schedule member functions, called by problem_setup(), to read 
//  some input files and use the information to setup the schedule.

#include "workforc.h"

//----------------------------------------------------------
//  Read SCHD_nnn.WF for last week and setup the posible fills
//  for the emp_days in the week before the one to be scheduled.

//  Note that if we get a shift we don't recognize, we just use
//  shift[0].  This is not ideal, but at least we know the emp
//  did work that day.

void  schedule::read_lweek (void)
{
  char        buf[31], msg[200];
  int         iemp, iday, ishft;
  empl_day   *ed;

  wlog ("   Reading the schedule file from the previous week...\n");

  strcpy (msg, "\nYou must have a schedule for the previous week\n");
  strcat (msg, "before WorkForce can make a schedule for this one.\n");
  strcat (msg, "Make the week-number one less and click \"Schedule\".\n");
  strcat (msg, "Abort.\n");

  char filnam[20];
  sprintf (filnam, "SCHD_%3.3d.WF", int(week_num-1));
  wif lweek_wif (filnam, msg);
  while (lweek_wif.get()) {

    istrstream istr(lweek_wif.rec,strlen(lweek_wif.rec));
    if (!istr) {
      wlog ("Error creating string stream from record: \n");
      wlog2 (lweek_wif.rec, "\n  ABORT.\n");
      wfexit(1);
    }

    strcpy (buf, get_s(&istr,20));
    iemp  = emp_name_to_index (buf);
    if (iemp < 0) {
      wlog3 ("......Ex-employee from last week ignored: ", buf, "\n");
      continue;
    }

    for (iday=0; iday<Week_Start; iday++) {

      strcpy (buf, get_s(&istr,6));
      ishft  = shft_abbr_to_index (buf);
      if (ishft < -1) ishft = 0;

      ed = emp_array[iemp].day + iday;    // The employee-day.
      ed->new_pos_fill (ishft, Ok, Ok);

    }   // End of loop through days of the previous week.

  }   // End of loop through fill records.

  lweek_wif.close();

}

//----------------------------------------------------------
//  Read AVAIL.WF and setup the posible fills for each empl_day
//  in the week to be scheduled.
//
//  The line for "Days" must be before shift lines for each emp.

void  schedule::read_avail (void)
{
  char        buf[21];
  int         iemp, iday, ishft;
  goodness    gs, gsd;
  empl_day   *ed;
  shft_day   *sd;

  wlog ("   Reading the Availabilities File...\n");

  wif det_wif ("AVAIL.WF");
  while (det_wif.get()) {

    istrstream istr(det_wif.rec,strlen(det_wif.rec));
    if (!istr) {
      wlog ("Error creating string stream from record: \n");
      wlog2 (det_wif.rec, "\n  ABORT.\n");
      wfexit(1);
    }

    iemp = emp_id_to_index (get_s(&istr,12));
    if (iemp < 0) {
      wlog ("Incorrect employee-id (or remark-line without\n");
      wlog ("exclamation-mark) on line: \n");
      wlog2 (det_wif.rec, "\n  ABORT.\n");
      wfexit(1);
    }

    strcpy (buf, get_s(&istr,4));
    if (!strcmp(buf,"Days")) {         // Line is for days of week...
      ishft = -1;
    } else {                           // Line is for a shift...
      ishft = shft_abbr_to_index (buf);
      if (ishft < 0) {
        wlog3 ("Incorrect shift-abbreviation \"", buf, "\"\n");
        wlog  ("or remark-line without exclamation-mark) on line: \n");
        wlog2 (det_wif.rec, "\n  ABORT.\n");
        wfexit(1);
      }
    }

    strcpy (buf, get_s(&istr,20));
    if ( ! ug2g (buf, &gs) ) {
      wlog3 ("Incorrect code for shift-goodness \"", buf, "\" on line: \n");
      wlog2 (det_wif.rec, "\n  ABORT.\n");
      wfexit(1);
    }

    for (iday=Week_Start; iday<Num_Days; iday++) {

      strcpy (buf, get_s(&istr,20));
      if ( ! ug2g (buf, &gsd) ) {
        wlog3 ("Incorrect status-code \"", buf, "\" on line: \n");
        wlog2 (det_wif.rec, "\n  ABORT.\n");
        wfexit(1);
      }

      ed = emp_array[iemp].day + iday;

      if (ishft < 0) {    // Goodness is for day of week...
        ed->set_g_day (gsd);
      } else {            // Goodness is for a shift...
        sd = shft_array[ishft].day + iday;
        if (ed->g_day > Imposible && sd->emps_req > 0 &&
            gsd > Imposible && !(gsd == Neutral && gs <= Imposible)) {
          ed->new_pos_fill (ishft, gs, gsd);
        }
      }

    }    // End of loop through days in current week.

  }   // End of loop through fill records.

  det_wif.close();

}

//----------------------------------------------------------
//  Read DAYSOFF.WF and apply the requests for days off which affect
//  the week to be scheduled.

void  schedule::read_daysoff (void)
{
  char        buf[31];
  int         iemp, iday;
  empl_day   *ed;

  wlog ("   Reading the Days-Off file...\n");

  wif doff_wif ("DAYSOFF.WF");
  while (doff_wif.get()) {

    istrstream istr(doff_wif.rec+1,strlen(doff_wif.rec));
    if (!istr) {
      wlog ("Error creating strstream from record: \n");
      wlog2 (doff_wif.rec, "\n  ABORT.\n");
      wfexit(1);
    }

    strcpy (buf, get_s(&istr,12));
    iemp = emp_id_to_index (buf);
    if (iemp < 0) {
      wlog3 ("Days-Off for ex-employee ignored:", buf, "\n");
      continue;
    }

    short int iw1 = get_i (&istr);
    if (iw1 < 1 || iw1 > 999) {
      wlog ("Start-Week must be between 1 and 999 on line:\n");
      wlog2 (doff_wif.rec, "\n  Abort\n");
      wfexit(1);
    }

    short int id1 = get_i (&istr);
    if (id1 < 1 || id1 > 7) {
      wlog ("Start-Day must be between 1 and 7 on line:\n");
      wlog2 (doff_wif.rec, "\n  Abort\n");
      wfexit(1);
    }

    short int iw2 = get_i (&istr);
    if (iw2 < iw1 || iw1 > 999) {
      wlog ("End-Week must be between Start-Week and 999 on line:\n");
      wlog2 (doff_wif.rec, "\n  Abort\n");
      wfexit(1);
    }

    short int id2 = get_i (&istr);
    if (id2 < id1 || id1 > 7) {
      wlog ("End-Day must be between Start-Day and 7 on line:\n");
      wlog2 (doff_wif.rec, "\n  Abort\n");
      wfexit(1);
    }

    if (week_num >= iw1 && week_num <= iw2) {
      if (week_num > iw1) id1 = 1;
      id1 = Week_Start + id1 - 1;
      if (week_num < iw2) id2 = 7;
      id2 = Week_Start + id2 - 1;
      fill  *f;
      for (iday=id1; iday<=id2; iday++) {
        ed = emp_array[iemp].day + iday;
        f = ed->pos_fill;
        while (f) {
          if ( ! f->day_off ) f->reject();
          f = f->next_fill;
        }
      }
    }

  }   // End of loop through records.

  doff_wif.close();

}

//----------------------------------------------------------
//  Read TIMEOFF.WF and apply the requests for time off which affect
//  the week to be scheduled.

void  schedule::read_timeoff (void)
{
  char        buf[31];
  int         iemp, iweek, iday, stime, etime;
  empl_day   *ed;

  wlog ("   Reading the Time-Off file...\n");

  wif toff_wif ("TIMEOFF.WF");
  while (toff_wif.get()) {

    istrstream istr(toff_wif.rec+1,strlen(toff_wif.rec));
    if (!istr) {
      wlog ("Error creating strstream from record: \n");
      wlog2 (toff_wif.rec, "\n  ABORT.\n");
      wfexit(1);
    }

    strcpy (buf, get_s (&istr,12));
    iemp = emp_id_to_index (buf);
    if (iemp < 0) {
      wlog3 ("Time-Off for ex-employee ignored: ", buf, "\n");
      continue;
    }

    iweek = get_i (&istr);
    if (iweek < 1 || iweek > 999) {
      wlog ("Week-Number must be between 1 and 999 on line:\n");
      wlog2 (toff_wif.rec, "\n  Abort\n");
      wfexit(1);
    }

    iday = get_i (&istr);
    if (iday < 1 || iday > 7) {
      wlog ("Day-Number must be between 1 and 7 on line:\n");
      wlog2 (toff_wif.rec, "\n  Abort\n");
      wfexit(1);
    }

    stime = get_i (&istr);
    if (stime < 0 || stime > 2400) {
      wlog ("Start-Time must be between 0 and 2400 on line:\n");
      wlog2 (toff_wif.rec, "\n  Abort\n");
      wfexit(1);
    }

    etime = get_i (&istr);
    if (etime < stime || etime > 2400) {
      wlog ("End-Time must be between Start-Time and 2400 on line:\n");
      wlog2 (toff_wif.rec, "\n  Abort\n");
      wfexit(1);
    }

    if (week_num == iweek) {
      fill  *f;
      ed = emp_array[iemp].day + iday;
      f = ed->pos_fill;
      while (f) {
        if ( ! f->day_off && f->shft->end_time   > stime &&
                             f->shft->start_time < etime) {
          f->reject();
        }
        f = f->next_fill;
      }
    }

  }   // End of loop through records.

  toff_wif.close();

}

//----------------------------------------------------------
//  Read ROTATION.WF and apply the rotations to the posible fills.

void  schedule::read_rotations (void)
{
  char        buf[31];

  wlog ("   Reading the Rotations file...\n");

  wif rot_wif ("ROTATION.WF");
  while (rot_wif.get()) {

    istrstream istr(rot_wif.rec+1,strlen(rot_wif.rec));
    if (!istr) {
      wlog ("Error creating strstream from record: \n");
      wlog2 (rot_wif.rec, "\n  ABORT.\n");
      wfexit(1);
    }

    strcpy (buf, get_s (&istr,12));
    short int iemp = emp_id_to_index (buf);
    if (iemp < 0) {
      wlog3 ("Rotation for ex-employee ignored: ", buf, "\n");
      continue;
    }

    strcpy (buf, get_s (&istr,12));
    short int rday = dow_to_iday (i_dow_start, buf);
    if (rday < 0) {
      wlog2 ("Bad day-of-week abbreviation \"", buf);
      wlog3 ("\" in rotation for ", emp_array[iemp].name, ".\n");
      wlog  ("Must be Mon, Tue, Wed, Thu, Fri, Sat or Sun.");
      wfexit(1);
    }

    char wk1_shft[21];
    strcpy (wk1_shft, get_s (&istr,20));

    char wk2_shft[21];
    strcpy (wk2_shft, get_s (&istr,20));

    char wk3_shft[21];
    strcpy (wk3_shft, get_s (&istr,20));

    goodness r_g;
    char rot_fail_g[21];
    strcpy (rot_fail_g, get_s (&istr,20));
    if ( ! ug2g (rot_fail_g, &r_g) ) {
      wlog3 ("Incorrect status-code \"", rot_fail_g, "\" on line: \n");
      wlog2 (rot_wif.rec, "\n  ABORT.\n");
      wfexit(1);
    }

    short int i_rot_week = get_i (&istr);

    char this_shft[21];
    if (i_rot_week == 1) {
      strcpy (this_shft, wk1_shft);
    } else if (i_rot_week == 2) {
      strcpy (this_shft, wk2_shft);
    } else if (i_rot_week == 3) {
      strcpy (this_shft, wk3_shft);
    } else {
      wlog ("\"This is week #\" must be between 1 and 3 on line:\n");
      wlog2 (rot_wif.rec, "\n  Abort\n");
      wfexit(1);
    }

    empl_day *ed = emp_array[iemp].day + rday;
    fill *f = ed->pos_fill;
    while (f) {
      f->apply_rotation (this_shft, r_g);
      f = f->next_fill;
    }

  }   // End of loop through records.

  rot_wif.close();

}

//----------------------------------------------------------
