//  Brian Marshall      bmarshal@agt.net
//......................................

//==========================================================
//  WorkForce  header file.

//  In one run, WorkForce will attempt to produce a schedule for one
//  week.  The days in a schedule include the week to be scheduled
//  plus enough preceeding days that we can determine goodnesses of
//  the emps working the first day of the new week.

//  Shfts and emps in a schedule, and fills in an empl_day are implemented
//  as an array of elements, each of which is a member in a doubly-linked
//  ordered list.  The most important/hopeful is the top of the linked-list.
//  Array access can be used to loop through the objects, changing the key.

//  "Static Components" of an object are the attributes of the object.
//  They are set when the object is created.

//  "Dynamic Components" of an object vary as the problem is being solved.
//  They are the results of judgements.

//  "Structural Components" of an object implement the list to which the
//  object belongs in an owning object.

//  Dynamic components of an object are altered only by it's member
//  functions in response to particular events:
//    Event   Function             Called When
//    <ps>   problem_setup()     WorkForce is started
//    <ts>   try_setup()         I begin to try to complete a schedule
//    <ds>   day_setup()         Object becomes the "current day".
//    <df>   do_fill()           An employee works a shift on a day.
//    <uf>   undo_fill()         I backout a fill.
//    <sb>   save_bsf()          a "best so far" sched is found
//    <a>    note_avail()        empl_day says it can work this shft_day.
//    <b>    note_busy()         empl_day says it can't work this shft_day.
//    <t>    note_try()          I have finished a try.
//    <bh>   broaden_horizons    I have lost all hope for improvement

//  For an object owned by another, the functions above are called only
//  by the owner, except:
//     shft_day.note_avail()   called by   empl_day.try_setup()
//     shft_day.note_busy()    called by   empl_day.do_fill()

//  An object is "done" when has been satisfied in relation to building
//  a schedule.  The do_fill() function of an object sets 'done' when
//  the object can tell that the fill it has just done has done the object.
//  'done' is reset by undo_fill().

//  A "try" is an attempt to complete the schedule.  A "failure" occurs
//  when I realize that I will not be able to complete the schedule
//  (because I can't find an emp for a shft or a shft for an emp).  A
//  try is "finished" when I stop trying (either because the schedule
//  is done or because a failure has occured).

//  Badges are identifiers on classes which are used to ensure that
//  they are legit - that they are not the product of a bad pointer.

//--------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strstream.h>

#define     Badge_Id    407466

#define     True   int(1)
#define     False  int(0)

#define     null  0

typedef short int  rtn_stat;
#define     Success rtn_stat(0)
#define     Failure rtn_stat(1)

typedef long int  goodness;
#define	    Imposible   goodness(-99999)
#define	    Very_Bad    goodness(  -512)
#define	    Bad	        goodness(   -64)
#define	    Poor        goodness(    -8)
#define     Poorish     goodness(    -2)
#define	    Expected    goodness(    -1)
#define     Neutral     goodness(     0)
#define	    Ok	        goodness(     1)
#define	    Goodish     goodness(     2)
#define	    Good        goodness(     4)
#define     Very_Good   goodness(     8)
#define     Manditory   goodness( 99999)

typedef long int  importance;
#define     Irrelevant           importance(  0)
#define	    Trivial	         importance(  4)
#define	    Slightly_Important   importance(  8)
#define	    Somewhat_Important	 importance( 16)
#define	    Fairly_Important	 importance( 32)
#define	    Quite_Important	 importance( 64)
#define	    Very_Important	 importance(128)
#define	    Critical	         importance(256)

#define     Imp_Divsr   importance(32)

// goodness = (goodness * importance) / Importance_Divisor.

//--------------------------------------------------------
//  The limits.

#define  Max_Emp    40
#define  Max_Shft   15

#define  Week_Len   int( 7)
#define  Num_Days   int(14)
#define  Week_Start int( 7)

//--------------------------------------------------------
//  The classes.

class schedule;   // A staff-scheduling problem.
class shift;      // A requirement to have work done.
class shft_day;   // A day of a shift.
class employee;   // A person who can work shifts.
class empl_day;   // A day for an employee.
class fill;       // A shift an employee can work on a day.
class wif;        // A WorkForce Input File.

struct failure;    // What we tried but failed to do.

//--------------------------------------------------------
//  Schedule.

class schedule {    // A staff-scheduling problem.
//  Static Components
  int             week_num;       // User assigned number for week to do.
  char            title[61];      // Title for output schedule file.
  short int       i_dow_start;    // Index of day-of-week of i_week_start ("Mon"=0).
  long int        num_try;        // Number of times to try.
//  Shifts and Employees
  short int       num_shft;       // Number of shifts.
  shift          *shft_array;     // Beginning of array of shfts.
  shift          *shft;           // List of shfts, most important first.
  short int       num_emp;        // Number of employees.
  employee       *emp_array;      // Beginning of array of emps.
  employee       *emp;            // List of emps,  most important first.
//  Dynamic Components
  short int       itry;           // Number of current try.
  goodness        g;              // Goodness of current schedule.
  goodness        g_seniority;    // Goodness of seniority order.
  fill           *worst_fill;     // Worst fill in current try.
  goodness        g_prev;         // Goodness of schedule before this one.
  goodness        g_try;          // Success of last finished try.
  failure        *fail;           // Reason last try failed (or null).
  failure        *fail_prev;      // Reason try before last failed (or null).
  goodness        g_bsf;          // Goodness of "Best so Far" schedule.
  fill           *worst_bsf_fill; // Copy of worst_fill in bsf schedule.
  goodness        hope;           // Expectation that g_bsf can be improved.
  goodness        g_cutoff;       // No fills with this g or worse considered.
  short int       num_shft_done;  // Number of shfts done.
  short int       num_emp_done;   // Number of emps  done.
  short int       num_fail;       // Number of times try failed.
  short int       tries_since_bsf;// Number of tries since last best-so-far.
  short int       done;           // True: all shfts and emps done.
//  Friends
  friend class shift;
  friend class shft_day;
  friend class employee;
  friend class empl_day;
  friend class fill;
// Functions for Friends
  // Functions for problem_setup()
  void       rec_setup (char *dept_rec);
  void       read_lweek   (void);
  void       read_avail   (void);
  void       read_daysoff (void);
  void       read_timeoff (void);
  void       read_rotations (void);
  void       check_posibility (void);        // Solution is posible?
  void       check_integrity (void);
  // Functions for find_good_schedule()
  void       try_setup (void);         // Setup before a try.
  void       try_to_complete (void);   // Try to complete this schedule.
  void       save_bsf (void);          // Save as "best so far".
  void       backout  (void);          // Backout the schedule some.
  void       broaden_horizons (void);  // Prepare to try different stuff.
  void       log_failures (void);      // Write about failures to log.
  // Functions for try() and backout()
  fill      *pick_fill (void);
  void       do_fill   (fill*);
  void       note_try (void);
  void       other_g_aspects (void);
  void       det_g_seniority (void);
  void       undo_fill (fill*);
  void       write_schd (int first_day, int bsf_flag, FILE*);
  // General Functions.
  int        shft_abbr_to_index  (char *sname);
  int        emp_name_to_index   (char *ename);
  int        emp_id_to_index     (char *id);
  int        ug2g (char *user_g_code, goodness *g);
  char      *g2str (goodness g);
public:
  void       problem_setup (int week_num);
  int        find_good_schedule (void);      // Return true on success.
  void       write_bsf (void);               // Write best-so-far schedule.
  void       log_worst_bsf_fill (void);      // Write stats to log.
};

//--------------------------------------------------------
//  Shift.

class shift {      // A requirement to have work done.
  long int        badge;         // If authentic: badge == Badge_Id
// Static Components
  short int       ishft;         // Index into array of shifts in schd.
  char            name[5];       // Name for shift.
  short int       start_time;    // Start time (like 800).
  short int       end_time;      // Ending time (like 1600).
  short int       minutes;       // Length of shift in minutes.
//  Shift-days.
  shft_day       *day;           // Array of days.
//  Dynamic Components
  importance      imp;           // Importance.
  importance      i_been_prob;   // "Been a problem" imp factor.
  shft_day       *cur_day;       // First day not done (or last).
  short int       done;          // True if all days done.
  short int       num_fail;      // Number of times involved in failure.
//  Structural Components  <setup by schedule constructor>
  shift          *next_shft;     // Next most important shft in schd.
  shift          *prev_shft;     // Previous shft or null if most important.
  schedule       *schd;          // Schedule to which shift belongs.
//  Friends
  friend class schedule;
  friend class shft_day;
  friend class employee;
  friend class empl_day;
  friend class fill;
//  Functions for Friends
  void  problem_setup (char* rec);
  void  rec_setup (char* rec);   // Setup from shift file record.
  void  check_integrity (void);
  void  try_setup (void);        // Setup to attempt another completion.
  void  det_imp (void);          // Determine importance and re-position.
  void  do_fill   (fill*);       // Apply passed fill.
  void  note_try (void);         // A try has finished.
  void  undo_fill (fill*);       // Unapply passed fill.
  void  log_imp (void);          // Write why important to log.
  void  broaden_horizons (void); // Prepare to try different stuff.
};

class shft_day {    // A day of a shift.
  long int        badge;         // If authentic: badge == Badge_Id
//  Static Components
  short int       iday;          // Index into array of shft_days in shft.
  short int       emps_req;      // Number of emps required.
//  Dynamic Components
  short int       emps_could;    // Num emps which can work later.
  short int       emps_can;      // Num emps which can work now.
  short int       emps_have;     // Number of emps working.
  importance      i_been_prob;   // "Been a problem" imp factor.
  short int       done;          // True if emps_have == emps_req.
  short int       num_fail;      // Number of times involved in failure.
//  Structural Components
  shft_day       *next_day;      // The next day for this shft.
  shft_day       *prev_day;      // The previous day for this shft.
  shift          *shft;          // The shft to which this shft_day belongs.
//  Friends
  friend class schedule;
  friend class shift;
  friend class employee;
  friend class empl_day;
  friend class fill;
//  Functions for Friends
  void  problem_setup (short int a_emps_req);
  void  check_integrity (void);
  void  try_setup (void);
  void  do_fill (fill*);
  void  undo_fill (fill*);
  void  note_could (void); // An emp could when it gets to this day.
  void  note_can (void);   // An emp can now.
  void  note_busy (void);  // An emp which could is now busy.
  void  note_try (void);
  void  broaden_horizons (void);
};

//--------------------------------------------------------
//  Employee.

class employee {    // Someone who can work shifts.
  long int        badge;         // If authentic: badge == Badge_Id
//  Static Components
  short int       iemp;          // Index into array of emps in schd.
  char            id[13];        // Id number for employee.
  char            name[21];      // Name of employee.
  short int       days_per_week;  // Num days to work in week.
  short int       off_per_week;  // Num days off in week (derived).
  short int       min_per_week;  // Num minutes per week.
  goodness        g_ot;          // Goodness of working one  extra shft.
  goodness        g_ut;          // Goodness of working one  fewer shfts.
  short int       seniority;     // 1 (min) to 99 (max).
  importance      i_seniority;   // Seniority imp factor (derived).
  char            vary_shfts[2]; // Prefer varying shifts (Y/N).
// Employee-Days.
  empl_day       *day;           // Array of days.
//  Dynamic Components
  importance      imp;           // Importance of being done next.
  importance      i_left_bhnd;   // "Being left behind" imp factor.
  importance      i_been_prob;   // "Been a problem" imp factor.
  importance      i_strategic;   // "Grass is greener" style imp factor.
  importance      i_poor_shfts;  // "Has already got poor shifts" imp fac.
  empl_day       *cur_day;       // First day not filled.
  short int       done;          // True if all days done.
  goodness        g;             // Goodness in current week so far.
  goodness        g_prev;        // Goodness in previous try.
  goodness        g_try;         // How sucessful the last try was.
  goodness        g_bsf;         // Goodness in best-so-far schedule.
  short int       num_fail;      // Number of times involved in failure.
  shift          *shft_last;     // Last shift worked (today, if done).
  short int       min_ttp;       // Minutes this week "to this point".
  // "should work today?" Dynamic Components
  goodness        g_work_today;  // Dynamic goodness of working cur_day.
  goodness        g_off_today;   // Goodness of taking cur_day off.
  goodness        h_work_today;  // Hope of working cur_day.
  goodness        h_off_today;   // Hope of taking cur_day off.
  short int       days_can_work; // # days left can work.
  short int       days_can_off;  // # days left can off.
  short int       days_on;       // # days on  in a row.
  short int       days_off;      // # days off in a row.
  short int       days_on_need;  // # days still need to work.
  short int       days_off_need; // # days still need off.
// Structural Components
  employee       *next_emp;      // Next most important emp in schedule.
  employee       *prev_emp;      // Previous emp or null if most important.
  schedule       *schd;          // Schedule to which employee belongs.
//  Friends
  friend class schedule;
  friend class empl_day;
  friend class shift;
  friend class shft_day;
  friend class fill;
//  Functions for Friends
  void   problem_setup (char *rec);   // Setup before we begin.
  void   rec_setup (char* rec);  // Setup from shift file record.
  void   check_integrity (void);     // Check the problem setup.
  void   try_setup (void);       // Setup before a completion attempt.
  void   day_setup (void);       // Setup before a new day.
  void   do_fill   (fill*);      // Apply passed fill.
  void   note_try (void);        // A try has finished.
  void   undo_fill (fill*);      // Un-apply passed fill.
  void   save_bsf (void);        // Save situation as best-so-far.
  void   det_imp (void);         // Determine importance and adjust position.
  void   eval_work_today (void); // Determine g and h of working on cur_day.
  void   broaden_horizons (void);// Prepare to try different stuff.
  void   log_imp (void);         // Write why important to log.
  int    one_choice (void);      // True if only one choice available.
  int    must_work (void);       // True if must work to satisfy all shfts.
  void   det_i_poor_shfts (void);// Determine i_poor_shfts for emp.
};

class empl_day {    // A day for an employee.
  long int        badge;         // If authentic: badge == Badge_Id
//  Static Components
  short int       iday;          // Index of this day.
  goodness        g_day;         // Goodness associated with this day.
  short int       num_pos_fill;  // Number of posible fills.
  fill           *pos_fill_array;// Beginning of array of posible fills.
  fill           *pos_fill;      // List of posible fills in decreasing order of hope.
//  Dynamic Components
  fill           *cur_fill;      // The pos_fill to use.
  short int       done;          // True: cur_fill has been done.
  short int       can_work;      // There is an "on"  pos_fill.
  short int       can_off;       // There is an "off" pos_fill.
  importance      i_been_prob;   // "Been a problem" imp factor.
  importance      i_imp_fills;   // Has fills that look important.
  fill           *bsf_fill;      // The "Best so Far" pos_fill.
  goodness        g_bsf;         // Goodness of bsf_fill.
  short int       num_fail;      // Number of times involved in failure.
//  Structural Components
  empl_day       *next_day;      // The next day for this emp.
  empl_day       *prev_day;      // The previous day for this emp.
  employee       *emp;           // Employee to which this empl_day belongs.
//  Friends
  friend class schedule;
  friend class employee;
  friend class shift;
  friend class shft_day;
  friend class fill;
//  Functions for Friends
  void   problem_setup (void);   // Setup before we begin.
  void   check_integrity (void);     // Check the problem setup.
  void   try_setup (void);       // Setup before we attempt a completion.
  void   day_setup (void);       // Setup before we can consider this day.
  void   do_fill (fill*);        // Apply this fill to this empl_day.
  void   save_bsf (void);        // Save cur_fill as best-so-far fill.
  void   note_try (void);        // A try has finished.
  void   undo_fill (fill*);      // Un-apply this fill to this empl_day.
  void   broaden_horizons (void);// Prepare to try different stuff.
  void   new_pos_fill (short int ishft, goodness gs, goodness gsd);
  void   set_g_day (goodness);   // Set goodness of emp working this day.
};

//--------------------------------------------------------
//  Fill.     A step in a way  ("ways" are not defined in this version).

class fill {     // A shift an employee can work on a day.
  long int        badge;         // If authentic: badge == Badge_Id
//  Static Components
  short int      ifill;          // Index into array of posible fills.
  employee      *emp;            // The employee.
  short int      iday;           // The day.
  shift         *shft;           // A shft the emp can work.
  short int      day_off;        // True if this is a day off.
  goodness       g_static;       // Net static goodness of this fill.
  goodness       g_shft;         // Goodness associated with the shift.
  goodness       g_rotation;     // Goodness associated with rotations.
//  Dynamic Components which vary over the life of the problem.
  goodness       h_wrks_well;    // "This works well" hope factor.
  short int      has_worked;     // True if been part of Ok sched.
  short int      has_worked_emp; // True if been part of sched Ok for emp.
  goodness       h_been_prob;    // "Has been problems" hope factor.
  goodness       h_grss_grnr;    // "Grass is greener" hope factor.
  char           h_grss_grnr_x;  // 0 = halve h_grss_grnr next time.
//  Dynamic Comonents defined only on current day.
  goodness       g;              // Goodness (affect on schedule).
  goodness       g_shft_chg;     // Goodness of change from prev shft.
  goodness       g_work_today;   // Goodness of working or not.
  goodness       hope;           // Expectation that this is good thing to do.
  goodness       g_mpw;          // G: emp's minutes-per-week.
  goodness       h_mpw;          // Hope: emp's minutes-per-week.
//  Structural Components
  fill          *next_fill;      // Next most hopeful fill for empl_day.
  fill          *prev_fill;      // Previous fill or null if most hopeful.
  empl_day      *emp_day;        // Emp-day to which this fill belongs or null.
//  Friends
  friend class schedule;
  friend class employee;
  friend class empl_day;
  friend class shift;
  friend class shft_day;
//  Functions for Friends
  void  problem_setup (short int ishft, goodness gs, goodness gsd);
  void  check_integrity (void);  // Check the problem setup.
  void  day_setup (void);        // Setup before we can consider on it's day.
  void  do_fill (void);          // Apply affect of doing this fill.
  void  undo_fill (void);        // Unapply effect of doing this fill.
  void  note_try (void);         // A try including this fill has finished.
  void  det_g (void);            // Determine goodness of this fill.
  void  det_hope (void);         // Determine hope and move in list.
  void  broaden_horizons (void); // Prepare to try different stuff.
  void  reject (void);           // Don't consider doing this fill.
  void  log_hope (void);         // Write why hopefull to log.
  void  det_g_shft_chg (void);   // Det g of this considering yesterday.
  void  evaluate_mpw (void);     // Det g and h re minutes-per-week.
  void  apply_rotation (char *shft_str, goodness g_if_fail);
};

//  The goodness and hope of a fill are not meaningful unless
//  the fill is for the cur_day of the emp (ie. the first
//  day for which a cur_fill is not set).

//  Future: Doing something well which is quite different than what
//  we were doing before could cause h_grss_grnr to go up again.

//--------------------------------------------------------
//  WorkForce Input File.

class wif {     // A WorkForce Input File.
  char           fnam[20];       // File name.
  FILE          *ifp;            // The input file stream.
public:
  char           rec[201];       // Last record read from file.
   wif (char *fnam);             // Constructor: Open file.
   wif (char *fnam, char *msg);  // Constructor: Open file. msg on error.
  ~wif (void);                   // Destructor: close file.
  int  get (void);               // Get next data record and put in rec.
  void close (void);             // Close the file.
};

//--------------------------------------------------------
//  Failure.

//  Either I was looking for a shft for an emp, or I was looking for
//  an emp for a shft, and I failed.  One of emp and shft will be null.

struct failure {
  long int       badge;          // If authentic: badge == Badge_Id
  short int      iday;           // The day.
  shift         *shft;           // The shift or null.
  employee      *emp;            // The employee or null.
};

//--------------------------------------------------------
//  i/o and general functions.

int   open_log (void);
void  set_echo_log (int iflag);
void  wlog  (char *str);
void  wlog2 (char *str1, char *str2);
void  wlog3 (char *str1, char *str2, char *str3);
void  wlogi (int);
void  wlogl (long int);
FILE *log_file_ptr (void);

void  abort (char *msg);

long  int  get_l (istrstream *);
short int  get_i (istrstream *);
char*      get_s (istrstream *, int len);

int  all_white (char*);    // Return true if all white-space.

void  upcase (char*);      // Convert letters to upper case.

char *day_of_week (short int i_dow_start, short int iday);
char *day_of_wk   (short int i_dow_start, short int iday);

short int dow_to_iday (short int i_dow_start, char* dow_str);

void wfexit (int code);    // Write message and exit with code.

int   hr_minutes (char* hours);  // Return minutes in hours string.
char *min_hours (int minutes);   // Returns hours string from minutes.


//========================================================
